# Copyright (C) 2014 Association of Universities for Research in Astronomy, Inc.
# Contact: mschirme@gemini.edu
#  
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software 
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# $Id: band_def_UI.tcl,v 1.5 2013/02/04 18:38:05 gmmps Exp $

itcl::class gmmps::band_def_UI {
    inherit util::TopLevelWidget

    #  constructor

    constructor {catClass catname addXSIZE addTILT addPX addPY pixs bw resultz config_file det_img det_spec ra_img dec_img instType bdfile} {
	global vsDebugLevel
	global EXTRABANDSEP
	# extra separation between bands of two pixels
	set EXTRABANDSEP 0
	set vsDebugLevel 1
	set args {}
	eval itk_initialize $args
	set itk_option(-catalog) $catname
	set itk_option(-xsize) $addXSIZE
	set itk_option(-tilt) $addTILT
	set itk_option(-posx) $addPX
	set itk_option(-posy) $addPY
	set itk_option(-pixs) $pixs
	set itk_option(-resultz) $resultz
	set itk_option(-bw) $bw
	set itk_option(-bdfile) $bdfile
	set itk_option(-instType) $instType
	set config_file_ $config_file
	set catClass_ $catClass
	set DET_IMG_ $det_img
	set DET_SPEC_ $det_spec
	set RA_IMG_ $ra_img
	set DEC_IMG_ $dec_img
	
	set versionx 0.0?
	set home_ $::env(GMMPS)
	catch {set ver [open $home_/config/version.dat r]}
	catch {set versionxx [gets $ver versionx] }
	catch {::close $ver }
    }

    # destructor

    destructor {
        close
    }
    
    #############################################################
    #  Name: debugStatement
    #
    #  Description:
    #   Only print the passed in message, if the debug flag is set.
    #   BIG WARNING:  need to make this its own class.
    #############################################################
    #############################################################

    protected proc debugStatement {line} {
	global vsDebugLevel

	if { $vsDebugLevel == 1 } {
	    puts $line
	}
    }

    #############################################################
    #  Name: init
    #
    #  Description:
    #############################################################
    #############################################################

    protected method init {} {
	wm title $w_ "Nod And Shuffle Parameters ($itk_option(-number))"
	wm iconname $w_ "N&S Definition ($itk_option(-number))"
	set home $::env(GMMPS)
	
	# Populate defaultBandHt_ and defaultBandHtNames_ lists.
	setDefaultBandHt 
	
	#debugStatement "band ht defaults:\n$bandHtDefault_\n$bandHtDefaultNames_"
	
	make_layout $itk_option(-catalog) $itk_option(-xsize)

    }

    #############################################################
    #  Name: make_layout
    #
    #  Description: 
    #    Draw the spoc box.
    #############################################################
    #############################################################

    protected method make_layout {mycatname addXSIZE} {

        add_menubar $mycatname
        add_widgets $mycatname $addXSIZE
        add_buttons $mycatname $addXSIZE
        add_progress_bar
        make_short_help
        
        update idletasks
        
        # Try to load previous N&S mode from config file, and switch to that 
        # mode if appropriate.
        set prevShuffleMode [config_get_values "shuffle_mode" ]
       	if {$prevShuffleMode == "bandShuffle"} {
	    global shuffleMode
	    set shuffleMode "bandShuffle"
	    
	    $w_.option.bandShuffle select
	    switchShuffleMode "startup"
	    #validateBandDef "startup"
	    calcBandAry
	    
	} else {
	    #validateBandDef "startup"
        }
    }


    #############################################################
    #  Name: add_menubar
    #
    #  Description:
    #    Draw menubar
    #############################################################
    #############################################################

    protected method add_menubar {mycatname } {
	TopLevelWidget::add_menubar

	set m [add_menubutton File "Display File menu"]

	add_menuitem $m command "Open Band Def File" \
	    {Open Stored Band Definition} \
	    -command [code $this loadBandDef]

	add_menuitem $m command "Save Band Def File" \
	    {Save Band Definition} \
	    -command [code $this saveBandDefAs]

	add_menuitem $m command "Close" \
	    {Close this window} \
	    -command [code $this cancel $mycatname ]
	
    }


    #############################################################
    #  Name: add_widgets
    #
    #  Description:
    #    Draw widgets
    #############################################################
    #############################################################

    protected method add_widgets {mycatname addXSIZE } {
	global EXTRABANDSEP
	# dialog widgets
	#        pack [LabelChoice $w_.option.bias \
	    #                -text "Code Method:" \
	    #                -value Normal_Opt \
	    #                -relief groove \
	    #                -orient Vertical \
	    #                -choice {Normal_Opt Max_Opt}] \
	    #                -side left -anchor w


	#debugStatement "add_widgets: bandWidth=$itk_option(-bw)"


	set banddef_fname $itk_option(-bdfile)

        pack [LabelMessage $w_.bandfile \
		  -text "Band Def Filename:" \
		  -value [file tail $banddef_fname ] \
		  -relief groove] \
	    -side top \
	    -anchor e \
	    -fill x
        
	pack [frame $w_.option -borderwidth 2 -relief raised] \
	    -side top -fill x

	pack [radiobutton $w_.option.microShuffle \
		  -text "Microshuffle Mode" \
		  -variable shuffleMode \
		  -value "microShuffle" \
		  -command [code $this switchShuffleMode]] \
	    -side top -anchor w

	pack [radiobutton $w_.option.bandShuffle \
		  -text "Band Shuffling Mode" \
		  -variable shuffleMode \
		  -value "bandShuffle"  \
		  -command [code $this switchShuffleMode]] \
	    [frame $w_.option.bandModeFrame]\
	    -side top -anchor w
	pack [button $w_.calcBands \
		  -text "Recalculate Bands" \
		  -command [code $this calcBandAry]] \
	    [button $w_.clearBands \
		 -text "Hide Bands" \
		 -command [code .skycat1.image.imagef.canvas delete bandrects]] \
	    [button $w_.defaults \
		 -text "Restore Defaults" \
		 -command [code $this restoreDefaults ]] \
	    -side left -anchor w -in $w_.option.bandModeFrame
	

	$w_.option.microShuffle select
	
	global band_tbl

	set band_tbl [util::vmTableList $w_.btbl \
			  -title "Shuffle Bands" \
			  -height 10 \
			  -layoutcommand [code $this set_show_cols] \
			  -headings {"Band ID" "Y Pos" "Band Size"} \
			  -headinglines 1 \
			  -sizes {10 30 10} \
			  -selectmode extended]

	pack  $band_tbl -fill x -side top -anchor s

	global image2
	global binning
	set ccdsum [image2 fits get CCDSUM]

	foreach {binx biny} $ccdsum {}

	set binning $binx

	if { $binx != $biny } {
	    tk_messageBox -message "WARNING Binning is Asymetric!!!\n\nCCDSUM=$ccdsum\nUsing binning=$binning" -parent $w_
	}
	
	# Search for stored values in .gmmps.cfg
	set slitLen [config_get_values "slitlen_ms" ]
	if {[lindex $slitLen 0 ] == ""} {
	    set slitLen 5.0
	}
	
	set shuffleAmt [config_get_values "shuffleamt_ms" ]
	if {[lindex $shuffleAmt 0 ] == ""} {
	    set shuffleAmt 5.0
	}
	
	#puts "shuffleamt: .$shuffleAmt."
	
	set defpx [expr $shuffleAmt * $binning / $itk_option(-pixs)]
	set defpx [format "%.0f" [expr ceil($defpx) ]]
	pack \
	    [LabelEntry $w_.option.nsSlitlenOrBandHt\
		 -text "Slitlen (arcsec):" \
		 -labelwidth 25  \
		 -value $slitLen \
		 -state disabled \
		 -command [code $this calcBandAry] \
		 -valuefont -adobe-courier-medium-r-*-*-*-120* \
		 -labelfont -Adobe-helvetica-bold-r-normal-*-12* ] \
	    -side top

	$w_.option.nsSlitlenOrBandHt.entry config -disabledforeground grey50

	pack [LabelEntry $w_.option.nodAmtOryOffset \
		  -text "Define Nods In Obs. Tool:" \
		  -labelwidth 25  \
		  -value 0 \
		  -command [code $this calcBandAry] \
		  -valuefont -adobe-courier-medium-r-*-*-*-120* \
		  -labelfont -Adobe-helvetica-bold-r-normal-*-12* ] \
	    -side top

	$w_.option.nodAmtOryOffset config -state disabled

	pack [LabelEntry $w_.option.shuffleAmt \
		  -text "Shuffle Amt (arcsec):" \
		  -labelwidth 25  \
		  -value $shuffleAmt \
		  -command [code $this calcShufflePx] \
		  -valuefont -adobe-courier-medium-r-*-*-*-120* \
		  -labelfont -Adobe-helvetica-bold-r-normal-*-12* ] \
	    [LabelEntry $w_.option.shufflePx \
		 -text "Shuffle Amt (unbinned pix):" \
		 -labelwidth 25  \
		 -value $defpx \
		 -command [code $this checkShufflePx] \
		 -valuefont -adobe-courier-medium-r-*-*-*-120* \
		 -labelfont -Adobe-helvetica-bold-r-normal-*-12* ] \
	    -side top

	# Bind KeyRelease event that will automatically update arcsec shuffle amount
	# into the pixel field. 
	bind $w_.option.shuffleAmt.entry <KeyRelease> [code $this calcShufflePx]
	bind $w_.option.shufflePx.entry  <KeyRelease> [code $this calcShuffleAmt]
	
	
	set catShuffSiz ""
	set catBandSiz ""
	set catBand1Y ""
	set catNodSiz ""
	#debugStatement "cat=$itk_option(-catalog)"	
	catch {
	    set catShuffSiz [exec grep "#fits SHUFSIZE" $itk_option(-catalog) | cut -f2 -d= ]
	    set catShuffSiz [string trim $catShuffSiz]
	}
	#debugStatement "SHUFSIZE=$catShuffSiz"

	catch {
	    set catBandSiz [exec grep "#fits BANDSIZE" $itk_option(-catalog) | cut -f2 -d= ]
	    set catBandSiz [string trim $catBandSiz]
	}
	#debugStatement "BANDSIZE=$catBandSiz"

	catch {
	    set catBand1Y [exec grep "#fits BAND1Y"  $itk_option(-catalog) | cut -f2 -d= ]
	    set catBand1Y [string trim $catBand1Y]
	}
	#debugStatement "BAND1Y=$catBand1Y"

	catch {
	    set catNodSiz [exec grep "#fits NODSIZE"  $itk_option(-catalog) | cut -f2 -d= ]
	    set catNodSiz [string trim $catNodSiz]
	}
	#debugStatement "NODSIZE=$catNodSiz"

	global shuffleMode

	# if these are present in the cat file, set them as defaults
	if {$catShuffSiz != ""} {
	    if {$catBandSiz != ""} {
		set shuffleMode "bandShuffle"
		$w_.option.bandShuffle select
		switchShuffleMode
		set catyOff [expr $catBand1Y - $catShuffSiz - ($EXTRABANDSEP/2)]
		$w_.option.nsSlitlenOrBandHt config -state normal
		$w_.option.nsSlitlenOrBandHt config -value [format "%.0f" $catBandSiz] 
		$w_.option.nodAmtOryOffset config -value [format "%.0f" $catyOff]
		$w_.option.nodAmtOryOffset config -state normal
		$w_.option.shufflePx config -value [format "%.0f" $catShuffSiz ]
		calcShuffleAmt
		calcBandAry
	    } elseif {$catNodSiz != ""} {
		set shuffleMode "microShuffle"
		$w_.option.microShuffle select
		switchShuffleMode
		$w_.option.nsSlitlenOrBandHt config -value [format "%.3f" $catShuffSiz] -state disabled
		$w_.option.nodAmtOryOffset config -value 0 
		$w_.option.nodAmtOryOffset config -state disabled
		$w_.option.shufflePx config -value [format "%.0f" $catShuffSiz ]
		calcShuffleAmt
	    } 
	}
	
	# Push slitlength into shuffle amount. 
	if {$shuffleMode == "microShuffle"} {
	    set sra [$w_.option.shuffleAmt get]
	    $w_.option.nsSlitlenOrBandHt config -state normal
	    $w_.option.nsSlitlenOrBandHt config -value $sra
	    $w_.option.nsSlitlenOrBandHt config -state disabled
	}
    }


    #############################################################
    #  Name: checkCombo
    #
    #  Description:
    #    Check combination of wavelength, filter & grating.
    #############################################################
    #############################################################

    protected method checkCombo { bw } {

        set Filter [$w_.option.filter get]
        set Grate [$w_.option.grateorder get]
	set itk_option(-bw) $bw
	#debugStatement "checkComb: bandWidth=$itk_option(-bw), grate=$Grate, filter=$Filter "

	if { $itk_option(-bw) < 350 || $itk_option(-bw) > 1050 } {
	    error_dialog "Central wavelength outside useful range for GMOS"
	} elseif { $Grate == "B1200" } {
	    if { $itk_option(-bw) > 595 } {
		error_dialog "Central wavelength too large, GMOS camera overfilled"
	    }
	} elseif { $Grate == "R831" && $itk_option(-bw) > 860 } {
	    error_dialog "Central wavelength too large, GMOS camera overfilled"
	}
    }


    protected method add_buttons {mycatname addXSIZE } {
	# dialog buttons
	pack [frame $w_.buttons -borderwidth 2 -relief raised] \
	    -side top -fill x

	pack [button $w_.spoc \
		  -text "Build N&S Mask Files" \
		  -command [code $this designMask]] \
	    [button $w_.close \
		 -text "Close" \
		 -command [code $this cancel $mycatname ]] \
	    -side left -expand 1 -pady 2m -in $w_.buttons

    }

    protected method add_progress_bar {} {
	# add a progress bar at the botton
	pack [ProgressBar $w_.progress \
		  -length 150] \
	    -side top -fill x
        $w_.progress.entry config -width 10
    }
    


    #############################################################
    #  Name: designMask
    #
    #  Description:
    #   This function is called when the "Design Mask" button
    #   is pressed. It creates the Master ODF file(s). 
    #       Calls the gmmps_spoc method.  Which will bring up
    #	     the popup to allow you to run the spoc algorithm.
    #
    #############################################################
    #############################################################

    public method designMask {} {
	set retv [validateBandDef ]
	if {$retv == "notValid"} {
	    error_dialog "Mask Definition invalid, cannot make N&S mask prototype."
	    return
	}
	
	#  Determine if slitpos_x and slitpos_y exist in the
	#  catalog files.
	
	set addPX ""
	set addPY ""
	
	catch {set addPX [exec grep slitpos_x $itk_option(-catalog) ]}
	catch {set addPY [exec grep slitpos_y $itk_option(-catalog) ]}
	
	if { $addPX == "" } {
	    set addPX "yes" 
	} else {
	    set addPX "no" 
	}
	if { $addPY == "" } {
	    set addPY "yes" 
	} else {
	    set addPY "no" 
	}
	
	# save the N&S configs
	saveBandDef
	
	
	set shuffleAmt [$w_.option.shuffleAmt get ]
	
	global shuffleMode
	if {$shuffleMode == "bandShuffle"} {
	    set isBandShuffle 1
	} else {
	    set isBandShuffle 0	    
	}
	
	#  Call the gmmps_spocxe .tcl init function
	if { [catch {utilReUseWidget cat::gmmps_spoc .skycat1.spoc $catClass_ $itk_option(-catalog) $itk_option(-xsize) $itk_option(-tilt) $addPX $addPY $itk_option(-pixs) $itk_option(-bw) $itk_option(-resultz) $itk_option(-instType) $config_file_ $DET_IMG_ $DET_SPEC_ $RA_IMG_ $DEC_IMG_ $itk_option(-bdfile) $isBandShuffle $shuffleAmt} msg]} {
	    debugStatement "Failed to spoc"
	    error_dialog "Mask Design popup already up."
	    puts "Mask Design popup detected... cannot open N&S Mask Design window."
	    return 
	}

    }

    protected method make_short_help {} {
	TopLevelWidget::make_short_help

	add_short_help $w_.widgets.width \
	    {set slit width (arcsec)}
	add_short_help $w_.widgets.mask \
	    {set number of ODF files (masks) to be generated}
	add_short_help $w_.option.filter \
	    {set filter to be used for the observation}
	add_short_help $w_.option.gratingorder \
	    {set grating&order to be used for the observation}
	add_short_help $w_.option.band \
	    {set central wavelength to be used for the observation}
	add_short_help $w_.label.catalog \
	    {name of input catalog}	    
	add_short_help $w_.frame \
	    {output of the Slit Positioning Optimization Code}	    
	add_short_help $w_.spoc \
	    {Make ODF: generate the slit catalog}
	add_short_help $w_.close \
	    {Close: close this window}
        add_short_help $w_.progress \
            {Progress bar: displays status of work in progress}
    }

    #########################################################################
    #  Name: cancel
    #
    #  Description:
    #     close
    #
    #########################################################################
    #########################################################################
    
    public method cancel {mycatname } {
	# This method gets called by anything that has a cancel button
	# on it..... not sure how to distinguish it.
	#  Need to pass back the current bandwith to vmAstroCat

	#	utilReUseWidget cat::vmAstroCat .skycat1.spoc $itk_option(-catalog) $itk_option(-xsize) $itk_option(-tilt) $itk_option(-bw) $this.results
        close
    }


    #########################################################################
    #  Name: close
    #
    #  Description:
    #     Destroy window
    #
    #########################################################################
    #########################################################################
    
    public method close {} {
    	# Clear bands on window close.
	set canvas .skycat1.image.imagef.canvas
    	catch {$canvas delete bandrects}
    	
        catch {destroy $w_ }
    }

    #########################################################################
    #  Name: switchShuffleMode
    #
    #  Description:
    #     called to switch shuffle modes
    #
    #########################################################################
    #########################################################################
    
    public method switchShuffleMode {args} {
	global shuffleMode
	
	# Save values to config file when switching modes.
	if {$shuffleMode == "bandShuffle"} {
	    config_save_values "microShuffle"
	} elseif {$shuffleMode == "microShuffle"} {
	    config_save_values "bandShuffle"
	}
	
	if {$args == "textonly"} {
	    if {$shuffleMode == "microShuffle"} {
		# micro shuffle mode entered
		$w_.option.nsSlitlenOrBandHt config \
		    -text "Slitlen (arcsec):" \
		    -state disabled
		
		$w_.option.nodAmtOryOffset config \
		    -text "Define Nods In Obs. Tool:"
		
	    } elseif {$shuffleMode == "bandShuffle"} {
		# band shuffle mode entered
		$w_.option.nsSlitlenOrBandHt config \
		    -text "Band Size (unbinned pix):" \
		    -state disabled
		
		$w_.option.nodAmtOryOffset config \
		    -text "Bands y Offset (unbinned pix):"
	    }
	} else {
	    if {$shuffleMode == "microShuffle"} {
		# micro shuffle mode entered
		
		# Search for stored values in .gmmps.cfg
		set slitLen [config_get_values "slitlen_ms" ]
		if {$slitLen == ""} {
		    #debugStatement "slitlen_ms not found. loading default"
		    set slitLen 5.0
		}
		set shuffleAmt [config_get_values "shuffleamt_ms" ]
		if {$shuffleAmt == ""} {
		    #debugStatement "shuffleamt_ms not found. loading default"
		    set shuffleAmt 5.0
		}
		
		# Edit values in widget.
		$w_.option.shuffleAmt config -value [format "%.3f" $shuffleAmt]
		calcShufflePx
		$w_.option.nsSlitlenOrBandHt config \
		    -text "Slitlen (arcsec):" \
		    -value $slitLen \
		    -state disabled
		$w_.option.nodAmtOryOffset config \
		    -text "Define Nods In Obs. Tool:" \
		    -value 0
		$w_.option.nodAmtOryOffset config -state disabled
		
	    } elseif {$shuffleMode == "bandShuffle"} {
		# band shuffle mode entered
		
		# Search for stored values in .gmmps.cfg
		set bandSize [config_get_values "bandsize_bs" ]
		if {$bandSize == ""} {
		    set bandSize [lindex $bandHtDefault_ 0 ]
		}
		set yOffset  [config_get_values "yoffset_bs" ]
		if {$yOffset == ""} {
		    set yOffset 0
		}
		set shufflePx [config_get_values "shufflepx_bs" ]
		if {$shufflePx == ""} {
		    set shufflePx [lindex $bandHtDefault_ 0 ]
		}
		
		# Edit values in widget.
		$w_.option.nsSlitlenOrBandHt config \
		    -text "Band Size (unbinned pix):" \
		    -value $bandSize \
		    -state normal
		$w_.option.shufflePx config -value [format "%.0f" $shufflePx ]
		$w_.option.nodAmtOryOffset config -state normal
		$w_.option.nodAmtOryOffset config \
		    -text "Bands y Offset (unbinned pix):" -value $yOffset
		
		calcShuffleAmt
	    }
	}
	
	if {$args != "startup"} {
	    calcBandAry
	}
    }

    #########################################################################
    #  Name: calcBandAry
    #
    #  Description:
    #     calculate bands, plot them (and then call saveBandDef)
    #
    #########################################################################
    #########################################################################
    
    public method calcBandAry {args} {
	global band_tbl
	global binning
	global image2
	global shuffleMode
	global EXTRABANDSEP
	
	set canvas .skycat1.image.imagef.canvas
	
	# first, check if micromode
	if {$shuffleMode == "microShuffle"} {
	    $band_tbl clear
	    $canvas delete bandrects
	    saveBandDef
	    return
	    
	} elseif {$shuffleMode == "bandShuffle"} {
	    # NOTE: this will use unbinned pixel coordinate space

	    # check if it's valid before we make a fool of ourself
	    #  drawing illegal bands that won't be allowed anyway
	    #  when gmMakeMasks is called.
	    if {$args != "novalidate"} {
		validateBandDef "calcBandAry"
	    }
	    
	    set target_im_ image2
	    set iHt [expr [$target_im_ height] * $binning]
	    set iWd [expr [$target_im_ width]  * $binning]
	    
	    # note: shufflePx is given in unbinned pixels
	    set shuffPx [$w_.option.shufflePx get]
	    set bandHt [$w_.option.nsSlitlenOrBandHt get]
	    set yOffset [$w_.option.nodAmtOryOffset get]
	    
	    if {$yOffset < 0} {
		set yOffset 0
		$w_.option.nodAmtOryOffset config -value 0
	    }
	    # the maximum coordinate
	    set maxY [expr $iHt - $bandHt - $shuffPx - ($EXTRABANDSEP/2) + 1]
	    
	    $band_tbl clear
	    
	    # clear the bands
	    $canvas delete bandrects
	    
	    #calc step size x,y for hash mark drawing
	    set ystep [expr ($shuffPx / $binning) / 10 ]
	    set xstep $ystep
	    
	    # Draw horizontal n&s lines.
	    set j 0
	    set lastyend 0
	    for {set y [expr $shuffPx + ($EXTRABANDSEP/2) + $yOffset + 1]} {$y <= $maxY} {set y [expr $y + $shuffPx + $bandHt + $EXTRABANDSEP]} {
		incr j
		
		#draw hatch mark of prohibited area
		set x0loc 0
		set x1loc [expr $iWd / $binning]
		set y0loc [expr $lastyend / $binning]
		set y1loc [expr $y / $binning]
		
		set lastyend [expr $y + $bandHt]
		
		#image2 convert coords 0 [expr $y / $binning] image xleft ybot canvas
		#image2 convert coords [expr $iWd / $binning] [expr ( $y + $bandHt ) / $binning] image xright ytop canvas
		image2 convert coords $x0loc $y0loc image xleft ybot canvas
		image2 convert coords $x1loc $y1loc image xright ytop canvas
		
		# draw the excluded rectangle
		$canvas create line \
		    $xleft $ybot \
		    $xright $ybot \
		    $xright $ytop \
		    $xleft $ytop \
		    $xleft $ybot \
		    -fill red \
		    -width 1 \
		    -tags bandrects
		
		# draw hashmarks row wise
		
		# set xstep [expr 20 * $binning / $itk_option(-pixs)]
		for { set xv $x0loc} { $xv < $x1loc} { set xv [expr $xv + $xstep]} {
		    image2 convert coords $xv $y0loc image xleft ybot canvas
		    image2 convert coords $xv $y1loc image xright ytop canvas
		    $canvas create line \
			$xleft $ybot $xright $ytop -fill red -tags bandrects
		} 
		
		for { set yv $y0loc} { $yv < $y1loc} { set yv [expr $yv + $ystep]} {
		    image2 convert coords $x0loc $yv image xleft ybot canvas
		    image2 convert coords $x1loc $yv image xright ytop canvas
		    $canvas create line \
			$xleft $ybot $xright $ytop -fill red -tags bandrects
		} 
	    }
	    
	    # draw one extra hash mark block from last band to top
	    #draw hatch mark of prohibited area
	    set x0loc 0
	    set x1loc [expr $iWd / $binning]
	    set y0loc [expr $lastyend / $binning]
	    set y1loc [expr $iHt / $binning]
	    
	    set lastyend [expr $y + $bandHt]
	    
	    #image2 convert coords 0 [expr $y / $binning] image xleft ybot canvas
	    #image2 convert coords [expr $iWd / $binning] [expr ( $y + $bandHt ) / $binning] image xright ytop canvas
	    image2 convert coords $x0loc $y0loc image xleft ybot canvas
	    image2 convert coords $x1loc $y1loc image xright ytop canvas
	    
	    # draw the excluded area rectangle
	    $canvas create line \
		$xleft $ybot \
		$xright $ybot \
		$xright $ytop \
		$xleft $ytop \
		$xleft $ybot \
		-fill red \
		-width 1 \
		-tags bandrects
	    
	    # draw hashmarks row wise
	    # set xstep [expr 20 * $binning / $itk_option(-pixs)]
	    for { set xv $x0loc} { $xv < $x1loc} { set xv [expr $xv + $xstep]} {
		image2 convert coords $xv $y0loc image xleft ybot canvas
		image2 convert coords $xv $y1loc image xright ytop canvas
		$canvas create line \
		    $xleft $ybot $xright $ytop -fill red -tags bandrects
	    } 
	    
	    for { set yv $y0loc} { $yv < $y1loc} { set yv [expr $yv + $ystep]} {
		image2 convert coords $x0loc $yv image xleft ybot canvas
		image2 convert coords $x1loc $yv image xright ytop canvas
		$canvas create line \
		    $xleft $ybot $xright $ytop -fill red -tags bandrects
	    } 
	    
	    set j 0
	    set lastyend 0
	    for {set y [expr $shuffPx + $yOffset + ($EXTRABANDSEP/2) + 1]} {$y <= $maxY} {set y [expr $y + $shuffPx + $bandHt + $EXTRABANDSEP]} {
		incr j
		$band_tbl add_row "band$j $y $bandHt"
		
		#draw band rect after so it's on top
		set bx0loc 0
		set bx1loc [expr $iWd / $binning]
		set by0loc [expr $y / $binning]
		set by1loc [expr ($y + $bandHt) / $binning]
		
		image2 convert coords $bx0loc $by0loc image xleft ybot canvas
		image2 convert coords $bx1loc $by1loc image xright ytop canvas
		
		# draw the band rectangle
		$canvas create line \
		    $xleft $ybot \
		    $xright $ybot \
		    $xright $ytop \
		    $xleft $ytop \
		    $xleft $ybot \
		    -fill yellow \
		    -width 1 \
		    -tags bandrects
	    }
	    
	    $band_tbl calculate_format
	}
	saveBandDef "" "calcBandAry"
    }


    #########################################################################
    #  Name: drawBands
    #
    #  Description:
    #     does the actual drawing of the bands. 
    #
    #########################################################################
    #########################################################################

    public proc drawBands {parent_window odfFile} {
	set home [pwd ]
	set canvas .skycat1.image.imagef.canvas
	set EXTRABANDSEP 0
	set target_im_ image2
	
	set bdfile ""
	catch {set bdfile [string trim [exec grep "#fits BDFILE" $odfFile | cut -f2 -d= ]]}
	
	if {$bdfile != "" && [string index $bdfile 0 ] == "'"} {
	    # Must be encased in ' ... ', remove. 
	    set bdfile [string trim [string range $bdfile 1 end-1 ]]
	}
	
	if {$bdfile != "" && [file dirname $bdfile ] == "."} {
	    # Make sure bdfile is a complete file path.

	    set bdfile [format "%s/%s" [file dirname $odfFile ] $bdfile ]
	}

	# if no band definition file can be found that is associated with the
	# given odf, try to load a tmp.bands file.
	if {$bdfile == ""} {
	    set tmpFileName [format "%s/tmp.bands" [file dirname $odfFile ]]
	    set tmpFileName2 [format "%s/tmp.bands" [pwd ]]
	    if {[file exists $tmpFileName ]} {
		set bdfile $tmpFileName
		tk_messageBox -message "Drawing bands from $tmpFileName.\nThis is a temporary file and may not reflect the actual band locations for this ODF." -parent $parent_window -title "Using tmp.bands"
	    } elseif {[file exists $tmpFileName2 ]} {
		set bdfile $tmpFileName2
		tk_messageBox -message "Drawing bands from $tmpFileName.\nThis is a temporary file and may not reflect the actual band locations for this ODF." -parent $parent_window -title "Using tmp.bands"

		#info_dialog "Drawing bands from $tmpFileName2.\nThis is a temporary file and may not reflect the actual band locations for this ODF."

	    } else {
		error_dialog [format "No band definition file or bands.tmp found in [file dirname $odfFile ].\n\nIf you are attempting to display the bands for a band shuffle mask design please be sure this dircectory contains an associated band definition file or copy of tmp.bands." [file dirname $odfFile ]]
		return 1
	    }
	}
	
	# If the file doesn't exist
	if {[file exists $bdfile] == 0} {
	    puts "Error: Cannot plot bands! \nThe band definition file $bdfile listed in your ODF catalog does not exist.\n\n"
	    error_dialog [format "Cannot plot bands!\n The band definition file $bdfile listed in your ODF catalog does not exist.\n"]
	    return 1
	}

	if {[catch {set shuffMode [exec grep "shuffleMode:" $bdfile | cut -f2 -d: ]} msg]} {
	    error_dialog [format "Cannot plot bands: $msg.\nshuffleMode not present in $itk_option(-bdfile)\nPlease re-run Design Nod & Shuffle Mask with correct parameters." ]
	    return 1
	} else {
	    set shuffleMode2 [string trim $shuffMode ]

	    if {$shuffleMode2 == "microShuffle"} {
		error_dialog "Cannot plot bands,\nN&S design window was last in microShuffle mode."
		return 1
	    }
	}
	
	# Find band height.
	if {[catch {set bandHt [string trim [exec grep "slitlen:" $bdfile | cut -f2 -d: ]]} msg]} {
	    error_dialog "Cannot plot bands: $msg.\nslitlen not present in $itk_option(-bdfile)\nPlease re-run Design Nod & Shuffle Mask with correct parameters."
	    return 1
	}	    
	
	# Find shuffPx.
	if {[catch {set shuffPx [string trim [exec grep "shufflePx:" $bdfile | cut -f2 -d: ]]} msg]} {
	    error_dialog "Cannot plot bands: $msg.\nshufflePx not present in $itk_option(-bdfile)\nPlease re-run Design Nod & Shuffle Mask with correct parameters."
	    return 1
	}	    
	
	# Find yoffset.
	if {[catch {set yOffset [string trim [exec grep "yOffset:" $bdfile | cut -f2 -d: ]]} msg]} {
	    error_dialog "Cannot plot bands: $msg.\nyOffset not present in $itk_option(-bdfile)\nPlease re-run Design Nod & Shuffle Mask with correct parameters."
	    return 1
	}

	# Find binning.
	if {[catch {set binning [string trim [exec grep "binning:" $bdfile | cut -f2 -d: ]]} msg]} {
	    error_dialog "Cannot plot bands: $msg.\nbinning not present in $itk_option(-bdfile)\nPlease re-run Design Nod & Shuffle Mask with correct parameters."
	    return 1
	}

	set iHt [expr [$target_im_ height] * $binning]
	set iWd [expr [$target_im_ width]  * $binning]
	
	
	# the maximum coordinate
	set maxY [expr $iHt - $bandHt - $shuffPx - ($EXTRABANDSEP/2) + 1]
	
	#$band_tbl clear
	
	# clear the bands
	$canvas delete bandrects

	#calc step size x,y for hash mark drawing
	set ystep [expr ($shuffPx / $binning) / 10 ]
	set xstep $ystep

	set j 0
	set lastyend 0
	for {set y [expr $shuffPx + ($EXTRABANDSEP/2) + $yOffset + 1]} {$y <= $maxY} {set y [expr $y + $shuffPx + $bandHt + $EXTRABANDSEP ]} {
	    incr j
	    
	    #draw hatch mark of prohibited area
	    set x0loc 0
	    set x1loc [expr $iWd / $binning]
	    set y0loc [expr $lastyend / $binning]
	    set y1loc [expr $y / $binning]

	    set lastyend [expr $y + $bandHt]

	    #image2 convert coords 0 [expr $y / $binning] image xleft ybot canvas
	    #image2 convert coords [expr $iWd / $binning] [expr ( $y + $bandHt ) / $binning] image xright ytop canvas
	    image2 convert coords $x0loc $y0loc image xleft ybot canvas
	    image2 convert coords $x1loc $y1loc image xright ytop canvas

	    # draw the excluded rectangle
	    $canvas create line \
		$xleft $ybot \
		$xright $ybot \
		$xright $ytop \
		$xleft $ytop \
		$xleft $ybot \
		-fill red \
		-width 1 \
		-tags bandrects

	    # draw hashmarks row wise

	    # set xstep [expr 20 * $binning / $itk_option(-pixs)]
	    for { set xv $x0loc} { $xv < $x1loc} { set xv [expr $xv + $xstep]} {
		image2 convert coords $xv $y0loc image xleft ybot canvas
		image2 convert coords $xv $y1loc image xright ytop canvas
		$canvas create line \
		    $xleft $ybot $xright $ytop -fill red -tags bandrects
	    } 

	    for { set yv $y0loc} { $yv < $y1loc} { set yv [expr $yv + $ystep]} {
		image2 convert coords $x0loc $yv image xleft ybot canvas
		image2 convert coords $x1loc $yv image xright ytop canvas
		$canvas create line \
		    $xleft $ybot $xright $ytop -fill red -tags bandrects
	    } 
	}

	# draw one extra hash mark block from last band to top
	#draw hatch mark of prohibited area
	set x0loc 0
	set x1loc [expr $iWd / $binning]
	set y0loc [expr $lastyend / $binning]
	set y1loc [expr $iHt / $binning]
	
	set lastyend [expr $y + $bandHt]
	
	#image2 convert coords 0 [expr $y / $binning] image xleft ybot canvas
	#image2 convert coords [expr $iWd / $binning] [expr ( $y + $bandHt ) / $binning] image xright ytop canvas
	image2 convert coords $x0loc $y0loc image xleft ybot canvas
	image2 convert coords $x1loc $y1loc image xright ytop canvas
	
	# draw the excluded area rectangle
	$canvas create line \
	    $xleft $ybot \
	    $xright $ybot \
	    $xright $ytop \
	    $xleft $ytop \
	    $xleft $ybot \
	    -fill red \
	    -width 1 \
	    -tags bandrects

	# draw hashmarks row wise
	# set xstep [expr 20 * $binning / $itk_option(-pixs)]
	for { set xv $x0loc} { $xv < $x1loc} { set xv [expr $xv + $xstep]} {
	    image2 convert coords $xv $y0loc image xleft ybot canvas
	    image2 convert coords $xv $y1loc image xright ytop canvas
	    $canvas create line \
		$xleft $ybot $xright $ytop -fill red -tags bandrects
	} 
	
	for { set yv $y0loc} { $yv < $y1loc} { set yv [expr $yv + $ystep]} {
	    image2 convert coords $x0loc $yv image xleft ybot canvas
	    image2 convert coords $x1loc $yv image xright ytop canvas
	    $canvas create line \
		$xleft $ybot $xright $ytop -fill red -tags bandrects
	} 


	set j 0
	set lastyend 0
	for {set y [expr $shuffPx + ($EXTRABANDSEP/2) + $yOffset ]} \
	    {$y <= $maxY} \
	    {set y [expr $y + $shuffPx + $bandHt + $EXTRABANDSEP]} {
		incr j
		#$band_tbl add_row "band$j $y $bandHt"
		
		#draw band rect after so it's on top
		set bx0loc 0
		set bx1loc [expr $iWd / $binning]
		set by0loc [expr $y / $binning]
		set by1loc [expr ($y + $bandHt) / $binning]

		image2 convert coords $bx0loc $by0loc image xleft ybot canvas
		image2 convert coords $bx1loc $by1loc image xright ytop canvas

		# draw the band rectangle
		$canvas create line \
		    $xleft $ybot \
		    $xright $ybot \
		    $xright $ytop \
		    $xleft $ytop \
		    $xleft $ybot \
		    -fill yellow \
		    -width 1 \
		    -tags bandrects
	    }
	
	return 0
    }


    #########################################################################
    #  Name: calcShufflePx
    #
    #  Description:
    #     calculate shufflePx from shuffleAmt
    #
    #########################################################################
    #########################################################################
    
    public method calcShufflePx {args} {
	global binning
	global shuffleMode
	
	set sra [$w_.option.shuffleAmt get]
	# note: shuffle pixels are always UNBINNED!, thus binning is 
	#  factored in below
	
	if { [catch {set spx [expr $sra * $binning / $itk_option(-pixs)]} ]} {
	    # invalid entry handing.
	    
	    $w_.option.shufflePx config -value "###"

	    # For microshuffle also push slitlen value to "Slitlen (arcsec):"
	    if {$shuffleMode == "microShuffle"} {		
		$w_.option.nsSlitlenOrBandHt config -state normal
		$w_.option.nsSlitlenOrBandHt config -value "###"
		$w_.option.nsSlitlenOrBandHt config -state disabled
	    }
	    
	} else {
	    # Valid entry...convert to pixels and push.
	    
	    set spxf [ expr ceil($spx) ] 
	    
	    $w_.option.shufflePx config -value [format "%.0f" $spxf ]
	    
	    # For microshuffle also push slitlen value to "Slitlen (arcsec):"
	    if {$shuffleMode == "microShuffle"} {		
		$w_.option.nsSlitlenOrBandHt config -state normal
		$w_.option.nsSlitlenOrBandHt config -value $sra
		$w_.option.nsSlitlenOrBandHt config -state disabled
	    }
	    
	}
    }


    #########################################################################
    #  Name: calcShuffleAmt
    #
    #  Description:
    #     calculate shuffleAmt from shufflePx
    #
    #########################################################################
    #########################################################################
    
    public method calcShuffleAmt {args} {
	global binning
	global shuffleMode	
	
	if {[catch {set spa [expr ceil([$w_.option.shufflePx get]) * $itk_option(-pixs) / $binning ]}]} {
	    $w_.option.shuffleAmt config -value "###"
	    
	    
	    # For microshuffle also push slitlen value to "Slitlen (arcsec):"
	    if {$shuffleMode == "microShuffle"} {		
		$w_.option.nsSlitlenOrBandHt config -state normal
		$w_.option.nsSlitlenOrBandHt config -value "###"
		$w_.option.nsSlitlenOrBandHt config -state disabled
	    }			
	} else {
	    # Valid entry...convert to arcseconds and push.
	    $w_.option.shuffleAmt config -value [format "%.3f" $spa]
	    
	    # For microshuffle also push slitlen value to "Slitlen (arcsec):"
	    if {$shuffleMode == "microShuffle"} {		
		$w_.option.nsSlitlenOrBandHt config -state normal
		$w_.option.nsSlitlenOrBandHt config -value [format "%.3f" $spa]
		$w_.option.nsSlitlenOrBandHt config -state disabled
	    }
	}

    }


    #########################################################################
    #  Name: loadBandDef
    #
    #  Description:
    #     loadBandDef
    #
    #     saves band definition to the band def file (bdfile)
    #     essentially this saves all the 
    #########################################################################
    #########################################################################
    
    public method loadBandDef {args} {
    	global shuffleMode
	
	if {$args == ""} {
	    set retStr [tk_getOpenFile -filetypes {{N&S .bands}} -parent $w_ -title "Band Def File"]
	}
	
	if {$shuffleMode == "microShuffle"} {
	    $w_.option.nsSlitlenOrBandHt config -state normal
	}
	
	if {$retStr != ""} {
	    set bfile [open $retStr r]
	    while { [gets $bfile line] >= 0} {
		#debugStatement "bd:$line"
		switch -regexp [lindex $line 0] {
		    "shuffleMode:" {
			switch [lindex $line 1] {
			    bandShuffle {
				$w_.option.bandShuffle select
			    }
			    microShuffle {
				$w_.option.microShuffle select
			    }
			}
		    }
		    "slitlen:" {
			$w_.option.nsSlitlenOrBandHt config -value [format "%.3f" [lindex $line 1]]
		    }
		    "shufflePx:" {
			$w_.option.shufflePx config -value [format "%.0f" [lindex $line 1 ]]
		    }
		    "shuffleAmt:" {
			$w_.option.shuffleAmt config -value [format "%.3f" [lindex $line 1]]
		    }
		    "nodAmt:" {
			# line below commented out... nod amount not part of
			# mask definition now... define in Observing Tool 
			#$w_.option.nodAmtOryOffset config -value [lindex $line 1]
			$w_.option.nodAmtOryOffset config -value 0
		    }
		}
	    }
	    
	    if {$shuffleMode == "microShuffle"} {
		$w_.option.nsSlitlenOrBandHt config -state disabled
	    }
	    
	    switchShuffleMode textonly
	}
    }
    

    #########################################################################
    #  Name: saveBandDefAs
    #
    #  Description:
    #     saveBandDefAs
    #
    #     saves band definition to the band def file (bdfile)
    #     this opens a file save as dialog before calling saveBandDef
    #########################################################################
    #########################################################################
    
    public method saveBandDefAs {args} {
	set retStr [tk_getSaveFile -filetypes {{N&S .bands}} -parent $w_ -title "Band Def File"]
	if {$retStr != ""} {
	    saveBandDef "saveas" $retStr
	}
    }

    #########################################################################
    #  Name: saveBandDef
    #
    #  Description:
    #     saveBandDef
    #
    #     saves band definition to the band def file (bdfile) 
    #     essentially this saves all the band vars for use by gmMakeMask
    #########################################################################
    #########################################################################
    
    public method saveBandDef {{mode ""} {args ""}} {
	# save the band definition file to the -bdfile file
	global shuffleMode
	global band_tbl
	
	#debugStatement "saveBandDef $args"
	
	config_save_values $shuffleMode
	
	# Don't re-validate after
	if {$args != "calcBandAry" && $args != "novalidate"} {
	    validateBandDef
	}
	
	file delete -force $itk_option(-bdfile)
	set bdfile [open $itk_option(-bdfile) "w"]
	
	#debugStatement "$itk_option(-bdfile) opened for writing"
	
	if {[file tail $itk_option(-bdfile) ] != "tmp.bands"} {
	    puts $bdfile "OT:$itk_option(-catalog)"
	}
	
	puts $bdfile "shuffleMode: $shuffleMode"
	
	#write slitlen from microshuffle mode
	if {$shuffleMode == "microShuffle"} {
	    $w_.option.nsSlitlenOrBandHt config -state normal
	    set sOb  [$w_.option.nsSlitlenOrBandHt get]
	    $w_.option.nsSlitlenOrBandHt config -state disabled
	} else {
	    set sOb  [$w_.option.nsSlitlenOrBandHt get]
	}
	
	puts $bdfile "slitlen: $sOb"
	if { $shuffleMode == "bandShuffle" } {
	    puts $bdfile "#fits BANDSIZE= $sOb" 
	}
	
	#write binning
	global binning
	puts $bdfile "binning: $binning"
	
	#write shufflePx in unbinned pixels
	set spx [$w_.option.shufflePx get]
	puts $bdfile "shufflePx: $spx"
	puts $bdfile "#fits SHUFSIZE= $spx"
	
	#write shufflePx in unbinned pixels
	set spx [$w_.option.shuffleAmt get]
	puts $bdfile "shuffleAmt: $spx"
	
	if { $shuffleMode  == "microShuffle" } {
	    set nodAmt 0
	    puts $bdfile "nodAmt: $nodAmt"
	    puts $bdfile "#fits NODSIZE = $nodAmt"
	} elseif { $shuffleMode == "bandShuffle" } {
	    set bands [$band_tbl get_contents]
	    set i 0
	    foreach brow $bands {
		incr i
		puts $bdfile "band: $i $brow"
		set by [lindex $brow 1]
		puts $bdfile "#fits BAND${i}Y= $by"
	    }
	    puts $bdfile [format "yOffset: %s" [$w_.option.nodAmtOryOffset get ]]
	}
	
	::close $bdfile
    }

    #########################################################################
    #  Name: config_save_values
    #
    #  Description:
    #	Saves N&S widget values to the .gmmps.cfg personal configuration file.
    #########################################################################
    #########################################################################

    protected method config_save_values {shufMode} {
	
	if {$shufMode == "microShuffle"} {
	    $w_.option.nsSlitlenOrBandHt config -state normal
	    set slitLen [$w_.option.nsSlitlenOrBandHt get ]
	    $w_.option.nsSlitlenOrBandHt config -state disabled
	    
	    set shuffleAmt [$w_.option.shuffleAmt get ]
	    
	    #debugStatement "config_save_values: ms mode, saving : $slitLen $shuffleAmt"
	    
	    $catClass_ gmmps_config_ [list "shuffle_mode" "slitlen_ms" "shuffleamt_ms"] [list $shufMode $slitLen $shuffleAmt ]
	    
	} else {	
	    set bandSize [$w_.option.nsSlitlenOrBandHt get ]
	    set yOffset  [$w_.option.nodAmtOryOffset get ]
	    set shufflePx [$w_.option.shufflePx get ]
	    
	    #debugStatement "config_save_values: bs mode, saving : $bandSize $yOffset $shufflePx"
	    
	    $catClass_ gmmps_config_ [list "shuffle_mode" "bandsize_bs" "yoffset_bs" "shufflepx_bs"] [list $shufMode $bandSize $yOffset $shufflePx ]
	}
    }

    #########################################################################
    #  Name: config_get_values
    #
    #  Description:
    #	Fetches stored data from the config file, if it exists.
    #########################################################################
    #########################################################################

    protected method config_get_values {keys} {
	
	set vals [$catClass_ gmmps_config_ [concat $keys ]]
	
	if {[llength $vals ] == 0} {
	    return ""
	} else {
	    return $vals
	}
    }
    
    
    #########################################################################
    #  Name: restoreDefaults
    #
    #  Description:
    #	Restore default values to the N&S widget.
    #########################################################################
    #########################################################################	
    
    protected method restoreDefaults {args} {
	global shuffleMode
	
	if {$shuffleMode == "microShuffle"} {
	    $w_.option.nsSlitlenOrBandHt config -state normal
	    $w_.option.nsSlitlenOrBandHt configure -value 5.000
	    $w_.option.nsSlitlenOrBandHt config -state disabled
	    $w_.option.shuffleAmt configure -value 5.000
	    calcShufflePx
	} else {
	    set def [lindex $bandHtDefault_ 0 ]
	    #debugStatement "Resetting default to $def $bandHtDefault_"
	    $w_.option.nsSlitlenOrBandHt configure -value [format "%.0f" $def]
	    $w_.option.nodAmtOryOffset configure -value 0
	    $w_.option.shufflePx configure -value [format "%.0f" $def]
	    calcShuffleAmt
	    
	    if {$args != "calcBandAry"} {
		calcBandAry $args
	    }
	}
	
	saveBandDef "" $args
    }


    #########################################################################
    #  Name: getShuffleMode
    #
    #  Description:
    #	returns the value of the shufflemode variable.
    #########################################################################
    #########################################################################	

    public method getShuffleMode {} {
	global shuffleMode
	
	return $shuffleMode		
    }

    #########################################################################
    #  Name: validateBandDef
    #
    #  Description:
    #     validates band definition values.
    #########################################################################
    #########################################################################
    
    public method validateBandDef {args} {
	global shuffleMode
	global band_tbl
	
	if { $shuffleMode  == "microShuffle" } {
	    $w_.option.nsSlitlenOrBandHt config -state normal
	    set slitlen [$w_.option.nsSlitlenOrBandHt get ]
	    
	    set shufPx  [$w_.option.shufflePx get ]
	    set shufAmt [$w_.option.shuffleAmt get ]
	    
	    if { $slitlen < 0 } {
		$w_.option.nsSlitlenOrBandHt configure -value 0.000
		error_dialog "Slitlen must be > 0\n"
	    }
	    
	    if { $shufPx < 0 } {
		$w_.option.shufflePx configure -value 0
		$w_.option.shuffleAmt configure -value 0.000
		error_dialog "ShufflePx must be > 0\n"
	    }
	    
	    if {$shufPx < $slitlen} {
		# Slit length is less than shuffle amount. Invalid. 

		# Query user for action.				
		set choice [tk_dialog $w_.nserror "N&S Value Error" "Required: Shuffle Amt >= Slit Length" "" 0 "Set Shuffle Amt = $slitlen" "Set Slit Length = $shufAmt" "Restore Defaults"  ]

		if {$choice == 0} {
		    $w_.option.shuffleAmt configure -value [format "%.3f" $slitlen]
		    calcShufflePx
		} elseif {$choice == 1} {
		    $w_.option.nsSlitlenOrBandHt configure -value [format "%.3f" $shufAmt ]
		} else {
		    restoreDefaults $args
		}
	    }
	    
	    $w_.option.nsSlitlenOrBandHt config -state disabled
	    
	} elseif { $shuffleMode  == "bandShuffle" } {
	    global binning	
	    set target_im_ image2
	    set iHt [expr [$target_im_ height] * $binning]

	    set bHt [$w_.option.nsSlitlenOrBandHt get ]
	    set sPx [$w_.option.shufflePx get ]
	    set yOffset [$w_.option.nodAmtOryOffset get ]			
	    
	    # Set valid to false for loop entry.
	    set valid 0

	    # Now assume bands are valid, until otherwise notified.
	    set valid 1
	    
	    if { $bHt < 0 } {
		$w_.option.nsSlitlenOrBandHt configure -value 0.000
		error_dialog "Band Size must be > 0"
		#set valid 0
	    }
	    
	    if { $sPx < 0 } {
		$w_.option.shufflePx configure -value 0
		$w_.option.shuffleAmt configure -value 0.000
		error_dialog "Shuffle Amt (in pix) must be > 0"
	    }
	    
	    if {$bHt > $sPx} {
		# Band size is greater than shuffle amount. Invalid. 
		
		# Query user for action.				
		set choice [tk_dialog $w_.nserror "N&S Value Error" "Error: Band Size must be <= Shuffle Amt in Pix." "" 0 "Set Band Size = $sPx" "Set Shuffle Amt = $bHt" "Restore Defaults"]

		if {$choice == 0} {
		    $w_.option.nsSlitlenOrBandHt configure -value [format "%.0f" $sPx ]
		} elseif {$choice == 1} {
		    $w_.option.shufflePx configure -value [format "%.0f" $bHt ]
		    calcShuffleAmt
		} else {
		    restoreDefaults $args
		}
	    }
	    
	    # Don't let bands become too big... defaults are max sizes.
	    if {[expr $sPx * 2 + $bHt + $yOffset ] > $iHt} {
		restoreDefaults "novalidate"
		
		if {$yOffset > 0} {
		    tk_messageBox -message "Cannot apply offset: spectra pushed off detector." -parent $w_
		}
		
	    }
	    
	    # Check current shuffle amt value against default values for other
	    # detectors.
	    if {$sPx != [lindex $bandHtDefault_ 0 ]} {
		set cnt 1
		foreach bandDefault [lrange $bandHtDefault_ 1 end ] {
		    
		    if {$sPx == $bandDefault} {
			set ourName [lindex $bandHtDefaultNames_ 0]
			set ourValue [lindex $bandHtDefault_ 0 ]
			set badName [lindex $bandHtDefaultNames_ $cnt]
			
			if {$args == "startup"} {
			    $w_.option.shufflePx configure -value [format %.0f $ourValue ]
			    
			    $w_.option.nsSlitlenOrBandHt config -state normal
			    $w_.option.nsSlitlenOrBandHt configure -value [format %.0f $ourValue ]
			    
			    if {$shuffleMode == "microShuffle"} {
				$w_.option.nsSlitlenOrBandHt config -state disabled						
			    }
			    
			} else {
			    set titleText "Possible Default Confusion" 
			    set dialogText "Possible error: $sPx is the default shuffle amount for the $badName $itk_option(-instType) detector.\n\nYou are designing a mask for the $ourName detector (default value: $ourValue)."
			    set choice [tk_dialog $w_.defaultConflict $titleText $dialogText "" 0 "Use $ourName Default ($ourValue)" "Keep Current Value ($sPx)" ]
			    if {$choice == 0} {
				$w_.option.shufflePx configure -value [format %.0f $ourValue ]
				$w_.option.nsSlitlenOrBandHt config -state normal
				$w_.option.nsSlitlenOrBandHt configure -value [format %.0f $ourValue ]
				calcShuffleAmt
				
				if {$shuffleMode == "microShuffle"} {
				    $w_.option.nsSlitlenOrBandHt config -state disabled						
				}
			    }
			}
			break
		    }
		    incr cnt
		}
	    }
	    
	}
	
	return "valid"
    }
    
    
    #########################################################################
    #  Name: getDefaultBandHt
    #
    #  Description:
    #	Constructs a list of all default band heights for this instrument, 
    #	as defined for each detector. The default for the detector that this
    #	mask is being designed for is to be placed in index 0, the rest may be
    #	in any order.
    # 	Also populates a corresponding list of the detector names corresponding
    #	to the default values.
    #	Returns both these lists.
    #########################################################################
    #########################################################################
    
    protected method setDefaultBandHt {} {
	set detFile [open "$home_/config/Detectors.dat" "r" ]
	set bandHtDefault_ [list ]
	set bandHtDefaultNames_ [list ]
	
	# Gather detector information from config file. 
	# (See Detectors.dat for more info on config line order.)
	while {[gets $detFile line] >= 0} {
	    set line [split $line "\t" ]
	    set detInst [string trim [lindex $line 1 ]]
	    
	    # Instuments match so this is a relevant detector.
	    # NOTE GMOS-N == GMOS-S in this situation.
	    if {[string range $detInst 0 3 ] == [string range $itk_option(-instType) 0 3 ]} {
		set detId [string trim [lindex $line 2 ]]
		set detName [string trim [lindex $line 3 ]]
		set detBandSize [string trim [lindex $line 8 ]]
		
		# Compare loaded detector ID with target DET_SPEC_ (as was 
		# passed in to the constructor).
		if {$detId == $DET_SPEC_} {
		    #If they match insert the default at the start of the list.
		    set bandHtDefault_ [linsert $bandHtDefault_ 0 $detBandSize ]
		    set bandHtDefaultNames_ [linsert $bandHtDefaultNames_ 0 $detName ]
		} else {
		    # If they don't match append value.
		    set bandHtDefault_ [lappend bandHtDefault_ $detBandSize ]
		    set bandHtDefaultNames_ [lappend bandHtDefaultNames_ $detName ]
		}
	    }
	}
	
	::close $detFile
	return [list $bandHtDefault_ $bandHtDefaultNames_]
    }
    
    
    #########################################################################
    #  Name: checkShufflePx
    #
    #  Description: Format and recalculate the "shuffle amt (in pixels)" 
    #	entrybox.
    #########################################################################
    #########################################################################
    
    protected method checkShufflePx {args} {
	
	# Set value in shufflePx to ceiling. 
	$w_.option.shufflePx configure -value [expr ceil([$w_.option.shufflePx get ]) ]
	
	# Apply ceiling value to shuffle amt.
	calcShuffleAmt
	
	validateBandDef
	
	# Redraw bands. 
	calcBandAry		
    }
    

    # Contains the type of file loaded in, Object Tbl, Master OT , or fitsFile 
    itk_option define -catalog catalog Catalog "" 

    # Options indicating col's that exist
    itk_option define -xsize xsize XSize "no"
    itk_option define -tilt tilt Tilt "no"
    itk_option define -resultz resultz Resultz ""
    itk_option define -posx posx Posx ""
    itk_option define -posy posy Posy ""
    itk_option define -pixs pixs Pixs ""
    itk_option define -bw bw Bw "600"
    itk_option define -instType instType InstType ""

    # Name of the gmmps program configuration file. 
    protected variable config_file_

    # Reference to the catalog widget instance that called this widget.
    protected variable catClass_

    # Holds detector ID value for preimage.
    protected variable DET_IMG_ ""
    
    # Holds detector ID value for mask design.
    protected variable DET_SPEC_ ""

    # Holds the RA of the preimage
    protected variable RA_IMG_ ""
    
    # Holds the DEC of the preimage
    protected variable DEC_IMG_ ""

    protected variable home_ ""
    
    # A list of all possible default band heights. 
    protected variable bandHtDefault_ ""
    
    # A list of detector names corresponding to the defaults in $defaultBandHt_
    protected variable bandHtDefaultNames_ ""

    # QueryResult widget used to display search results
    #   protected variable results_ {}


    protected common nfilter
    protected common ngrate
    protected common ngtilt
    protected common buf_filter
    protected common buf_grate
    protected common buf_grequest
    protected common buf_gtilt
    protected common gTilt
    protected common anaMorphic
    protected common lmin
    protected common lmax
    protected common SpecLen
    protected common dpix
    protected common gRequest
    protected common versionx
    protected common banddef_fname
}
