/*
** Copyright (C) 2014 Association of Universities for Research in Astronomy, Inc.
** Contact: mschirme@gemini.edu
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software 
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*$Id: gmCat2Fits.c,v 1.2 2011/04/25 18:27:32 gmmps Exp $
 *
 ************************************************************************
 ****      D A O   I N S T R U M E N T A T I O N   G R O U P        *****
 *
 * (c) 2000                         (c) 2000
 * National Research Council        Conseil national de recherches
 * Ottawa, Canada, K1A 0R6          Ottawa, Canada, K1A 0R6
 * All rights reserved              Tous droits reserves
 * 					
 * NRC disclaims any warranties,    Le CNRC denie toute garantie
 * expressed, implied, or statu-    enoncee, implicite ou legale,
 * tory, of any kind with respect   de quelque nature que se soit,
 * to the software, including       concernant le logiciel, y com-
 * without limitation any war-      pris sans restriction toute
 * ranty of merchantability or      garantie de valeur marchande
 * fitness for a particular pur-    ou de pertinence pour un usage
 * pose.  NRC shall not be liable   particulier.  Le CNRC ne
 * in any event for any damages,    pourra en aucun cas etre tenu
 * whether direct or indirect,      responsable de tout dommage,
 * special or general, consequen-   direct ou indirect, particul-
 * tial or incidental, arising      ier ou general, accessoire ou
 * from the use of the software.    fortuit, resultant de l'utili-
 *                                  sation du logiciel.
 *
 ************************************************************************
 *
 * FILENAME
 * gmCat2Fits.c
 *
 * PURPOSE:
 *  Routine to translate Minimal ODF catalog file to fits.
 * 
 * PARAMETER
 *  <input>  Minimal ODF catalog input full file name (without extention)
 *  <output> Minimal ODF output fits file name (without extention)
 *  <N.Mask> number of masks to be prepared
 *
 * FUNCTION NAME(S)
 * main -			Main function entry.
 * writePadding -		Padd the table with spaces.
 * writeFitsInfo -		Write the fits file.
 *
 * Input Catalog <input.cat)
 *   It is composed by 13 columns:
 *   ID: object identification (INTEGER)
 *   RA: object RA in degrees (FLOAT)
 *   DEC: object DEC in degrees (FLOAT)
 *   x: object x-coordinate in pixels (FOV left down corner = 0,0) (FLOAT)
 *   y: object x-coordinate in pixels (FOV left down corner = 0,0) (FLOAT)
 *   specpos_x: spectrum center x-coordinate in pixels (FLOAT)
 *   specpos_y: spectrum center y-coordinate in pixesl (FLOAT)
 *   slitpos_x: slit center x-coordinate in arcsec (FLOAT)
 *   slitpos_y: slit center y-coordinate in arcsec (FLOAT)
 *   slitsize_x: slit x-length in pixels (FLOAT)
 *   slitsize_y: slit y-length in pixels (FLOAT)
 *   slittilt: object tilt
 *   mag : magnitude
 *   priority: object priority flag (CHAR)
 *   slittype: object type flag (CHAR)
 *
 * Input fits keywords file <input.cfg>, this is generated by gmmps_spoc.tcl.
 * Each line starts with # and is 80 char long and contains all additional
 * keywords to be written to the output Minimal ODF Fits file.
 *
 *  
 * Output FITS file File, binary table with following columns:
 *      "ID", "RA", "DEC", "x_ccd", "y_ccd", "specpos_x", "specpos_y",
 *      "slitpos_x","slitpos_y", "slitsize_x", "slitsize_y", "slittilt", 
 *      "MAG", "priority", "slittype" 
 *
 *
 *INDENT-OFF*
 * $Log: gmCat2Fits.c,v $
 * Revision 1.2  2011/04/25 18:27:32  gmmps
 * Forked from 0.401.12 .
 *
 * Revision 1.2  2011/01/25 22:44:50  gmmps
 * Wave_ccd .fits bug fixed.
 *
 * Revision 1.1  2011/01/24 20:02:12  gmmps
 * Compiled for RedHat 5.5 32 and 64 bit.
 *
 * Revision 1.3  2002/09/19 02:09:19  callen
 * minor change to comment
 *
 * Revision 1.2  2002/08/24 04:00:20  callen
 * the main change was in gmCat2Fits.c so that RA stored in degrees in cat files
 * is converted to hours in the FITS file which the astronomers prefer (as requested
 * buy Inger and Kathy).  The change in gmConvert2Cat.c is just a comment to point
 * out where the RA is multiplied by 15, which is the conversion from hours to degrees.
 *
 * Revision 1.1.1.1  2002/07/19 00:02:09  callen
 * importing gmmps as recieved from Jennifer Dunn
 * gmmps is a skycat plugin and processes for creating masks
 *
 * Revision 1.8  2001/12/11 00:07:24  dunn
 * Added the writing of 2 additional columns specpos_x and y.
 *
 * Revision 1.7  2001/11/27 23:04:55  dunn
 * Got rid of vltPort.h
 *
 * Revision 1.6  2001/11/08 19:59:44  dunn
 * Was reading 1 pass underlines, but there is no empty line.
 *
 * Revision 1.5  2001/11/01 23:16:14  dunn
 * Reset status when keyword put fails.
 *
 * Revision 1.4  2001/10/22 17:28:32  dunn
 * Reported fits errors, and got rid of potential tab from puts.
 *
 * Revision 1.3  2001/10/19 23:55:46  dunn
 * *** empty log message ***
 *
 * Revision 1.2  2001/10/19 23:54:30  dunn
 * Fixed it to read in and attempt to add keywords to final Minimal header.
 *
 * Revision 1.1  2001/10/11 21:03:19  dunn
 * Initial revision
 *
 *INDENT-ON*
 *
 ****      D A O   I N S T R U M E N T A T I O N   G R O U P        *****
 ************************************************************************
*/


/*
 *  Local Defines
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <ctype.h>
#include <fitsio.h>

/*
 *  Local Types
 */

#define NUM_FIELDS 16		/* Number of entries in following metaInfo */

char *ttype[NUM_FIELDS] = 
    { "ID", "RA", "DEC", "x_ccd", "y_ccd", "specpos_x", "specpos_y",
       "slitpos_x","slitpos_y", "slitsize_x", "slitsize_y", "slittilt", 
       "MAG", "priority", "slittype", "wave_ccd"};
char *tunit[NUM_FIELDS] =
/*    { "#", "deg", "deg","pixels", "pixels", "pixels", "pixels", "arcsec",
      @@cba: changed first "deg" to "H" to convert from degrees to hours for RA*/
    { "#", "H", "deg","pixels", "pixels", "pixels", "pixels", "arcsec", 
       	"arcsec", "arcsec", "arcsec", "deg", "#", "c",  "c", "pixels" };
char *tform[NUM_FIELDS] =
    { "1J", "1E", "1E", "1E", "1E", "1E", "1E", "1E", "1E", "1E", "1E", 
	"1E", "1E", "1A", "1A", "1E" };

   
/*
 *  Function Prototypes *** ALL LOCAL FUNCTIONS TO BE PROTOTYPED ***
 */

int writeFitsInfo( fitsfile *, char *, char *, char *);

/*
 *  Data Structures
 */


/*
 *  Macros
 */



/*
 ************************************************************************
 *+
 * FUNCTION: main
 *
 * RETURNS: int [success or failure]
 *
 * DESCRIPTION: Main entry point to function.
 *
 * [NOTES:]:
 *-
 ************************************************************************
 */

int main (int argc, char *argv[]) {
  /*
   * Internal variable declaration.
   */
  
  int n,l;
  int nexp;			/* Number of mask files.	*/
  int num;			/* Number of items in scanf.	*/
  int numAcq;              	/* Num of acqu. objects. 	*/
  char spocf[256],outrootf[256],inrootf[256];
  char fitsname[256];
  FILE	*spoc;			/* Input catalog file ptr.	*/
  fitsfile	*fits;		/* Output fits file ptr.	*/
  time_t	OurTime;
  int		status;		/* Function status.		*/
  char	creaTime[32];		/* Time created fits.		*/
  char	oneLine[256];		/* One line of input.		*/
  float	spec_x, spec_y;		/* Spectrum variables.		*/
  long	row;			/* row working on.		*/

  int id;			/* Id.				*/
  float ra;			/* Right assesion.		*/
  float dec;			/* Declanation.			*/
  float x;			/* X CCD posotion.		*/
  float y;			/* Y CDD Position.		*/
  float slitpos_x;		/* X offset pos.		*/
  float slitpos_y;		/* Y offset pos.		*/
  float slitsize_x;		/* X slit size.			*/
  float slitsize_y;		/* Y slit size.			*/
  float tilt;			/* Slit tilt.			*/
  float mag;			/* Magnitude.			*/
  char prior[2];		/* Need these to be 2 so that struct is 
				   correct size when writing to fits. */
  char type[2];			/* Slit type.			*/
  float wave_ccd;               /* distance to central wavelength from center of slit */

  char *tmp;
  char **tmpPtr;
  
  
  /*
   * COMMAND LINE INPUT 
   */
  printf("----------------------------------------------------\n");
  printf("      ASCII TABLE TO FITS CONVERSION PROGRAM\n");
  printf("----------------------------------------------------\n");
  printf("----------------------------------------------------\n");
  
  if (argc==4) {
    strcpy(inrootf,argv[1]);
    strcpy(outrootf,argv[2]);
    nexp=atoi(argv[3]);
    //printf("DEBUG: %s %s %s %s\n", argv[0], argv[1], argv[2], argv[3]);
  }
  else {
    printf("         WRONG INPUT COMMAND LINE: EXIT\n");
    printf("----------------------------------------------------\n");
    printf("PAR 1:Input catalog name with no extension\n");
    printf("PAR 2:Input spoc output file name, no extension and number\n");
    printf("PAR 3:Number of spoc output files to be converted\n\n");
    printf("Example: gmCat2Fits NGC1234_OT NGC1234_OTODF 2\n");
    printf("----------------------------------------------------\n");
    return 0;
  }
  
  /****** BIG WARNING, NOT REPORTING AN ERROR IF 1 OF THE MASK FILES IS NOT 
	  WRITTEN PROPERLY
  **********/
  
  /*
   *  For each mask file that exists...
   */
  
  for(l=1; l<=nexp; l++) {
    /*
     *  Create time string, and compose in/output file names,
     *  the MinimalODF cat  & MinimalODF Fits file.
     */

    OurTime = time(NULL);
    strftime(creaTime, sizeof(creaTime), "%Y-%m-%dT%H:%M:%S.000", 
	     localtime(&OurTime));
    
    sprintf(spocf,"%s%d.cat",outrootf,l);
    sprintf(fitsname,"!%s%d.fits",outrootf,l);
    
    /*
     *  Open input file mycatnameODF#.cat and read till
     *  you pass the dashes "-"
     */
    
    if((spoc = fopen(spocf, "r")) == NULL) {
	printf(" ERROR, can't open input file : %s.\n", spocf );
	continue;
    }
    while ( fgets( oneLine, 256, spoc) != NULL ) {
      if ( oneLine[0] == '-' && oneLine[1] == '-' )
	break;
    }
    
    /*
     *  Open and write the header units for the fits file. 
     */
    
    status = 0;
    if ( fits_create_file( &fits, fitsname, &status ) ) {
      printf(" ERROR, can't open %s, <%d>.\n", fitsname, status );
      continue;
    }

    if ( (status = writeFitsInfo( fits, fitsname, inrootf, creaTime)) != 0 ) {
      printf("Error writing the fits binary table, %s, <%d>.\n", 
	     fitsname,  status );
      fits_close_file( fits, &status );
      continue;
    }

    /*
     *  Read data from mycatname#.cat: id, ra, dec, slitxcenter, 
     *  slity center, slit x len, slit y len, d, mag, priority, type.
     */
    
    numAcq=n=row=0;
    while ( fgets( oneLine, 256, spoc) != NULL ) {
      
      /*
       *  Ignore lines beginning with comments, or less than 2 chars
       *  or that are too long and don't finish with a carriage return.
       */
      
      if ( oneLine[0] == '#' || strlen(oneLine) < 2 ) continue;
      if ( oneLine[ strlen(oneLine) - 1] != '\n' ) continue;
      
      /*
       *  Scan the line for the correct format.  If incorrect then don't
       *  continue.
       *  HUGE WARNING, all arrays are starting at 1..... must have
       *  been a fortran programmer..... 
       */

      if ( (num = sscanf( oneLine,"%d %f %f %f %f %f %f %f %f %f %f %f %f %s %s %f",
			  &id, &ra, &dec, &x, &y, &spec_x, &spec_y, &slitpos_x, &slitpos_y,
			  &slitsize_x, &slitsize_y, &tilt, &mag, prior, type, &wave_ccd))== EOF  ||
	   num != NUM_FIELDS ) {
	/*printf("sscanf failed, num=%d,\n", num );*/
	break;
      }
      row++;
      
      
      /*
       *  Save count of user-defined objects, and Acq. Objects.
       */
      
      prior[1] = '\0';
      type[1] = '\0';
      if (type[0] == 'A') printf(" Error, can't handle A type.\n");
      if (prior[0] == '0') numAcq++;
      
      /*
       *  Write the data.
       */
      status = 0;

      /*printf("About to write row=%ld, id = %d.\n", row, id );*/
      
      fits_write_col( fits, TINT, 1, row, 1, 1, &id, &status);
      /* @@cba:the .cat files use ra in degrees which they don't
	 want in the .fits file (for one it doesn't load) */
      ra /= 15; /* degrees to hours */
      fits_write_col( fits, TFLOAT, 2, row, 1, 1, &ra, &status);
      fits_write_col( fits, TFLOAT, 3, row, 1, 1, &dec,	&status);
      fits_write_col( fits, TFLOAT, 4, row, 1, 1, &x, &status);
      fits_write_col( fits, TFLOAT, 5, row, 1, 1, &y, &status);
      fits_write_col( fits, TFLOAT, 6, row, 1, 1, &spec_x, &status);
      fits_write_col( fits, TFLOAT, 7, row, 1, 1, &spec_y, &status);
      fits_write_col( fits, TFLOAT, 8, row, 1, 1, &slitpos_x, &status);
      fits_write_col( fits, TFLOAT, 9, row, 1, 1, &slitpos_y, &status);
      fits_write_col( fits, TFLOAT, 10, row, 1, 1, &slitsize_x, &status);
      fits_write_col( fits, TFLOAT, 11, row, 1, 1, &slitsize_y, &status);
      fits_write_col( fits, TFLOAT, 12, row, 1, 1, &tilt, &status);
      fits_write_col( fits, TFLOAT, 13, row, 1, 1, &mag, &status);
      tmp = prior;
      tmpPtr = &tmp;
      fits_write_col( fits, TSTRING, 14, row, 1, 1, tmpPtr, &status);
      tmp = type;
      tmpPtr = &tmp;
      fits_write_col( fits, TSTRING, 15, row, 1, 1, tmpPtr, &status);
      fits_write_col( fits, TFLOAT, 16, row, 1, 1, &wave_ccd, &status); 		    
      
      if ( status != 0 ) printf("Failed to write a column.\n");
      /*printf( "%d %f %f %f %f %f %f %f %f %f %f <%s> <%s>\n",
	id,ra,dec, x,y, slitpos_x,slitpos_y,
	slitsize_x,slitsize_y, tilt,mag, prior, type);*/
      
    }/* While reading in spoc file... */
    
    /*
     *  Close files.
     */
    
    fclose(spoc);
    fits_close_file( fits, &status );
    
  }/* For all mask files... */
  
  printf("            PROGRAM NORMALY TERMINATED\n");
  printf("----------------------------------------------------\n"); 
  return 0;
}

/*
************************************************************************
*+
* FUNCTION: 
* writeFitsInfo
*
* RETURNS: int [return status]
*
* DESCRIPTION: Write the fits primary header Info
*
* [NOTES:]:
*-
************************************************************************
*/

int writeFitsInfo
(
    fitsfile	*fp,		/* (in)  Fits File pter.	*/
    char 	*filename,	/* (in)  FITS Filename.		*/
    char 	*Infile,	/* (in)  Input Filename.	*/
    char	*creaTime	/* (in)  Time.			*/
)
{
  int	status;		        /* Return fwrite status.	*/
  int	hdrtype;	        /* Type of header unit.		*/
  char	*tmpFile;	        /* Temp file name.		*/
  char  hdrf[256];	        /* Fits hdr file.		*/
  FILE	*hdr;		        /* File ptr, ascii file of fits key*/
  char	oneLine[82];	        /* One line of input.		*/
  int	len;		        /* Line cnt.			*/
  char	newcard[FLEN_VALUE];	/* New keyword to write.	*/
  

  // Append a new empty binary table onto the FITS file 

  status = 0;
  if ( fits_create_tbl( fp, BINARY_TBL, 0, NUM_FIELDS, ttype, tform, 
			tunit, "MINIMAL.TAB", &status ) ) {
    printf("Failed to create binary table, <%d>.\n", status);
    fits_report_error( stderr, status );
    return( 1 );
  }

  // Compose input fits header info file name & open.
  
  sprintf(hdrf,"%s.cfg",Infile);
  if ((hdr = fopen(hdrf, "r")) == NULL) {
    printf("Can't open input fits keyword file: %s.\n", hdrf );
    return(0);
  }
  
  /* 
   * Read in 1 line at a time, skipping over keywords that we will 
   * be putting in like:
   * SIMPLE, END, BITPIX, NAXIS, FILENAME, FILE_OT, DATEMODF, BLANK, 
   * EXTEND, END.  Check that INSTRUME is given, otherwise just add 
   * something. Stop when we find END keyword, or nothing else to get.
   */

  while ( fgets( oneLine, 82, hdr) != NULL ) {
    status = 0;
    hdrtype = 0;
    
    //  Scan out the fits keyword, and see if we should write
    //  it to the fits file..

    if ( oneLine[0] != '#' || strlen(oneLine) < 16 ||
	 strncmp(oneLine, "#fits ", 6)!=0 ) continue;
    if ( strncmp( "SIMPLE",   oneLine, 6) == 0 ||
	 strncmp( "BITPIX",   oneLine, 6) == 0 ||
	 strncmp( "NAXIS",    oneLine, 5) == 0 ||
	 strncmp( "FILENAME", oneLine, 8) == 0 ||
	 strncmp( "FILE_OT",  oneLine, 7) == 0 ||
	 strncmp( "DATEMODF", oneLine, 8) == 0 ||
	 strncmp( "BLANK",    oneLine, 5) == 0 ||
	 strncmp( "EXTEND",   oneLine, 6) == 0  ) {
      // printf("Got one : %s\n", oneLine );
      continue;
    }
    
    //  Otherwise, parse the string and write the darn thing.
    
    len = strlen(oneLine);
    if ( oneLine[len-1] == '\n' ) oneLine[len-1] = '\0';
    len = strlen(oneLine);
    if ( oneLine[len-1] == '\t' ) oneLine[len-1] = '\0'; // found a tab
    len = strlen(oneLine);
    if ( oneLine[len-1] == 0x9 )  oneLine[len-1] = '\0'; // found a hex 9

    //printf("About to parse line: %s....\n", &oneLine[6]);
    fits_parse_template( &oneLine[6], newcard, &hdrtype, &status );
    if ( status != 0 || hdrtype != 0 ) {
      // Don't write this keyword.
      printf("Line did not parse: %s, <%d>\n", newcard, hdrtype);
      status = 0;
      continue;
    }
    fits_write_record(fp, newcard, &status );
    if ( status != 0 ) {
      fits_report_error(stderr, status );
      // printf("Failed to write <%s>\n\t<%d>, size=%d\n", newcard, status, strlen(newcard) );
      status = 0;
    }
  }
  fclose( hdr );

  //  Now write additional keywords:
  //  FILENAME
  //  TIME_ODF
  
  // str : filename
  status = 0;
  fits_write_key(fp, TSTRING, "FILENAME", (void *) Infile, NULL, &status );
  
  // str : file_ot
  if ( (tmpFile = strrchr( filename, '/')) != NULL ) tmpFile++;
  else tmpFile = filename;

  if ( status != 0 ) printf("Failed to write FILENAME\n");
  status = 0;
  fits_write_key(fp, TSTRING, "FILE_OT", (void *) tmpFile, NULL, &status );
    
  // str : Time of creation
  if ( (tmpFile = strrchr( creaTime, '/')) != NULL ) tmpFile++;
  else tmpFile = creaTime;
  
  if ( status != 0 ) printf("Failed to write FILE_OT\n");
  status = 0;

  fits_write_key(fp, TSTRING, "TIME_ODF", (void *) tmpFile, NULL, &status );
  if ( status != 0 ) printf("Failed to write TIME_ODF\n");

  return( status );
}
